/*
** Module   : AsyncRequestImpl.java
** Abstract : implementation of the async request resource.
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20130813 Created initial version.
*/

package com.goldencode.p2j.util;

import java.util.*;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.util.*;

/**
 * Implements the async request resource. Beside adding support for its attributes, the async
 * resource is also responsible of initiating the async call and raising the PROCEDURE-COMPLETE
 * event.
 * <p>
 * As some the mutable fields of this class can be modified by more than one thread, access to 
 * them is synchronized.
 */
public class AsyncRequestImpl
extends HandleChain
implements AsyncRequest,
           Runnable
{
   /** A handle to the SERVER resource. */
   private final handle server = new handle();
   
   /** The name of the EVENT-PROCEDURE. */
   private final character eventProcedure = new character();
   
   /** The context of the EVENT-PROCEDURE. */
   private final handle eventProcedureContext = new handle();
   
   /** The proxy procedure, if the remote call is for an internal procedure. */
   private final handle persistentProcedure = new handle();
   
   /** The name of the target procedure. */
   private final character procedureName = new character();
   
   /** The modes for the procedure's arguments. */
   private final String modes;
   
   /** The procedure's arguments. */
   private final Object[] args;
   
   /**
    * Flag indicating that the async call has completed (will be <code>false</code> if the call
    * was cancelled.
    */
   private boolean complete = false;
   
   /** Flag indicating that the async call has been cancelled. */
   private boolean cancelled = false;
   
   /** Flag indicating that the call was terminated with a QUIT condition. */
   private boolean quit = false;
   
   /** Flag indicating that the call was terminated with a STOP condition. */
   private boolean stop = false;

   /** Flag indicating if the resource is valid. */
   private boolean valid = true;

   /**
    * List of listeners which need to receive notifications when the call is started or finished.
    */
   private List<AsyncRequestListener> listeners = new ArrayList<AsyncRequestListener>();

   /** A worker indicating what will be invoked. */
   private Runnable request;
   
   /**
    * Create a new async request resource, by specifying the attributes related to the async
    * call definition.
    * 
    * @param    server
    *           The server used to invoke the remote call. May be a <code>SERVER</code> or the
    *           <code>SESSION</code> resource.
    * @param    request
    *           The request which will execute the procedure call.
    * @param    eventProcedure
    *           The event procedure name.
    * @param    eventProcedureContext
    *           The context of the event procedure.
    * @param    procedureName
    *           The name of the procedure to be invoked.
    * @param    persistentProcedure
    *           The handle to the proxy procedure to which the target procedure belongs. Will
    *           be unknown if this is not a remote call.
    * @param    modes
    *           The modes for the procedure's arguments.
    * @param    args
    *           The procedure's arguments.
    */
   public AsyncRequestImpl(handle    server,
                           Runnable  request,
                           character eventProcedure,
                           handle    eventProcedureContext,
                           character procedureName,
                           handle    persistentProcedure,
                           String    modes,
                           Object... args)
   {
      this.server.assign(server);
      this.request = request;
      this.eventProcedure.assign(eventProcedure);
      this.eventProcedureContext.assign(eventProcedureContext);
      this.persistentProcedure.assign(persistentProcedure);
      this.procedureName.assign(procedureName);
      this.modes = modes;
      this.args = args;
      
      setResourceType("ASYNC-REQUEST");
      setUnknown(false);
   }
   
   /**
    * Reports if this object is valid for use (has not been deleted).
    *
    * @return   <code>true</code> if we are valid (can be used).
    */
   @Override
   public boolean valid()
   {
      return valid;
   }

   /**
    * Returns a handle to the app server in which the procedure is ran.
    * 
    * @return   See above
    */
   @Override
   public handle getServerHandle()
   {
      return new handle(server);
   }

   /**
    * Get the <code>CANCELLED</code> attribute of this async request.
    * 
    * @return   See above.
    */
   @Override
   public synchronized logical isCancelled()
   {
      return new logical(cancelled);
   }

   /**
    * Get the <code>CANCELLED</code> attribute of this async request, as a Java boolean.
    * 
    * @return   See above.
    */
   public synchronized boolean _isCancelled()
   {
      return cancelled;
   }

   /**
    * Get the <code>COMPLETE</code> attribute of this async request.
    * 
    * @return   See above.
    */
   @Override
   public synchronized logical isComplete()
   {
      return new logical(complete);
   }

   /**
    * Get the <code>COMPLETE</code> attribute of this async request, as a Java boolean.
    * 
    * @return   See above.
    */
   public synchronized boolean _isComplete()
   {
      return complete;
   }

   /**
    * Get the <code>EVENT-PROCEDURE</code> attribute of this async request.
    * 
    * @return   See above.
    */
   @Override
   public character getEventProcedure()
   {
      return new character(eventProcedure);
   }

   /**
    * Get the <code>EVENT-PROCEDURE-CONTEXT</code> attribute of this async request.
    * 
    * @return   See above.
    */
   @Override
   public handle getEventProcedureContext()
   {
      return new handle(eventProcedureContext);
   }

   /**
    * Get the <code>PERSISTENT-PROCEDURE</code> attribute of this async request.
    * 
    * @return   See above.
    */
   @Override
   public handle getPersistentProcedure()
   {
      return new handle(persistentProcedure);
   }

   /**
    * Get the <code>PROCEDURE-NAME</code> attribute of this async request.
    * 
    * @return   See above.
    */
   @Override
   public character getProcedureName()
   {
      return new character(procedureName);
   }

   /**
    * Get the <code>QUIT</code> attribute of this async request.
    * 
    * @return   See above.
    */
   @Override
   public synchronized logical isQuit()
   {
      return new logical(quit);
   }

   /**
    * Get the <code>STOP</code> attribute of this async request.
    * 
    * @return   See above.
    */
   @Override
   public synchronized logical isStop()
   {
      return new logical(stop);
   }

   /**
    * Invoke the event procedure for this async request, if used.
    */
   public void run()
   {
      if (eventProcedure.isUnknown())
      {
         return;
      }

      List<Object> eventArgsList = new ArrayList<Object>();
      Object[] eventArgs = new Object[0];
      String eventModes = null;
      if (modes != null)
      {
         for (int i = 0; i < modes.length(); i++)
         {
            char mode = modes.charAt(i);
            if (SourceNameMapper.isOutputMode(mode))
            {
               eventArgsList.add(args[i]);
            }
         }
         
         eventModes = StringHelper.repeatChar(SourceNameMapper.INPUT_MODE, eventArgsList.size());
         eventArgs = eventArgsList.toArray();
      }
   
      // invoke the event procedure 
      try
      {
         SelfManager.pushSelf(new handle(this));
         ControlFlowOps.invokeInWithMode(eventProcedure,
                                         eventProcedureContext,
                                         eventModes,
                                         eventArgs);
      }
      finally
      {
         SelfManager.popSelf();
      }
   }

   /**
    * Wait for the remote request to finish.
    */
   synchronized void waitToFinish()
   {
      while (!(complete || cancelled))
      {
         try
         {
            this.wait();
         }
         catch (InterruptedException e)
         {
            break;
         }
      }
   }

   /**
    * Called when the async call has finished normally (not with a cancel).
    */
   synchronized void complete()
   {
      this.complete = true;
      notifyListeners(false);
   }
   
   /**
    * Called when the async call was cancelled by the user.
    */
   synchronized void cancelled()
   {
      this.cancelled = true;
      notifyListeners(false);
   }

   /**
    * Mark this async call as having ended with a STOP condition.
    */
   synchronized void setStop()
   {
      this.stop = true;
   }

   /**
    * Mark this async call as having ended with a QUIT condition.
    */
   synchronized void setQuit()
   {
      this.quit = true;
   }
   
   /**
    * Stop this async request by sending the STOP condition to the remote side. This will be
    * used only in cases when the async call is cancelled by the user.
    */
   synchronized void stop()
   {
      setStop();
      
      // TODO: can't duplicate in 4GL - i.e. the remote task is not cancelled, nor the STOP 
      // condition sent to the remote side.
   }

   /**
    * Register a new listener for notifications about the state of the async call.
    * 
    * @param    listener
    *           A listener instance.
    */
   void registerListener(AsyncRequestListener listener)
   {
      listeners.add(listener);
   }
   
   /**
    * Notify all listeners that the async call is about to start.
    */
   void notifyStart()
   {
      notifyListeners(true);
   }

   /**
    * Execute the async request. This will either raise the PROCEDURE-COMPLETE event or will
    * invoke the EVENT-PROCEDURE immediately.
    * 
    * @param    event
    *           Flag indicating if the PROCEDURE-COMPLETE event will be raised.
    */
   void execute(boolean event)
   {
      // execute the remote async call
      try
      {
         // TODO: how is RETURN-VALUE affected?
         request.run();
      }
      finally
      {
         if (!_isCancelled())
         {
            complete();
         }

         if (event)
         {
            ServerEvent serverEvent = new ServerEvent(handle.resourceId(this),
                                                      Keyboard.PROCEDURE_COMPLETE,
                                                      this);
            LogicalTerminal.postServerEvent(serverEvent);
         }
         else
         {
            // execute the event now
            run();
         }
         
         // notify that the request has finished.
         synchronized (this)
         {
            this.notify();
         }
      }
   }
   
   /**
    * The async resource can be deleted only if it has been completed or cancelled.
    */
   @Override
   protected synchronized boolean resourceDelete()
   {
      if (!(complete || cancelled))
      {
         final String msg = "Asynchronous request  has not completed.  Cannot DELETE";
         ErrorManager.recordOrThrowError(8978, msg);
         return false;
      }
      
      this.valid = false;
      return true;
   }
   
   /**
    * As this resource supports the name attribute in R/W mode, return <code>false</code>
    * 
    * @return   always <code>false</code>
    */
   @Override
   protected boolean hasNameReadOnly()
   {
      return false;
   }
   
   /**
    * Notify the listeners about the invocation request state (start or finish).
    * 
    * @param    start
    *           <code>true</code> to indicate invocation start, <code>false</code> to notify for
    *           invocation finish.
    */
   private void notifyListeners(boolean start)
   {
      for (AsyncRequestListener listener : listeners)
      {
         if (start)
         {
            listener.notifyStart();
         }
         else
         {
            listener.notifyFinish();
         }
      }
   }
}
